
# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.011.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "implementation of OpenID Connect for Plack::Auth::SSO",
  "dist_author" => [
    "Nicolas Franck, C<< <nicolas.franck at ugent.be> >>"
  ],
  "dist_name" => "Plack-Auth-SSO-OIDC",
  "dist_version" => "0.01",
  "license" => "perl",
  "module_name" => "Plack::Auth::SSO::OIDC",
  "recursive_test_files" => 1,
  "requires" => {
    "Crypt::JWT" => 0,
    "Data::UUID" => 0,
    "Data::Util" => 0,
    "Digest::SHA" => 0,
    "JSON" => 0,
    "LWP::UserAgent" => 0,
    "MIME::Base64" => 0,
    "Moo" => 0,
    "Plack" => 0,
    "Plack::Auth::SSO" => 0,
    "Try::Tiny" => 0,
    "URI" => 0,
    "perl" => "v5.10.1"
  }
);


my %fallback_build_requires = (
  "Module::Build" => "0.28"
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
